.bd-footer {
  font-size: 0.8rem;
}

html[data-theme="light"] {
  --pst-color-primary: hsl(222.2 47.4% 11.2%);
  --pst-color-secondary: #1774E5;
  --pst-color-secondary-bg: #1774E5;
  --pst-color-accent: #1774E5;
  --sd-color-secondary-highlight: #0062cc;
  --pst-color-shadow: rgba(0, 0, 0, 0.0);
}

html[data-theme="dark"] {
  --pst-color-primary: hsl(213 31% 91%);
  --pst-color-secondary: #017FFF;
  --pst-color-secondary-bg: #017FFF;
  --pst-color-accent: #017FFF;
  --sd-color-secondary-highlight: #0062cc;
  --pst-color-shadow: rgba(0, 0, 0, 0.0);
}

.bd-header-announcement {
  color: white;
}

.bd-header-announcement a {
  color: white;
}

.bd-header-announcement a:hover {
  color: white;
  text-shadow: 0.5px 0 0 currentColor;
}
nav.bd-links .current>a  {
  box-shadow: inset 1px 0 0 var(--pst-color-primary);
}
@media (forced-colors: active) {
  /* Top breadcrumbs navigation (ie: Home > Core > ...) */
  .bd-breadcrumbs .breadcrumb-item > a:focus-visible{
    border: 2px solid var(--pst-color-primary);
  }

  /* Left sidebar */
  nav.bd-links .navbar-nav .toctree-l1>a:focus-visible {
    border: 2px solid var(--pst-color-primary);
  }
  nav.bd-links .current>a  {
    box-shadow: none;
    border-left: 4px solid var(--pst-color-primary) !important;
  }

  /* Right sidebar */
  .bd-sidebar-secondary .sidebar-secondary-items .nav-item .active {
    box-shadow: none;
    border-left: 5px solid var(--pst-color-primary) !important;
  }
  .bd-sidebar-secondary .sidebar-secondary-items .nav-item>a:focus-visible {
    border: 2px solid var(--pst-color-primary);
  }
}
html[data-theme="light"] .bd-header {
  border-bottom: 1px solid var(--pst-color-border);
}

.admonition, div.admonition {
  border: 1px solid var(--pst-color-border);
}

.api-card {
  text-align: center;
  font-size: 1.2rem;
}

.api-card svg {
  font-size: 2rem;
}

.search-button-field {
  border-radius: var(--bs-btn-border-radius);
}

.bd-content .sd-tab-set .sd-tab-content {
  border: none;
  border-top: 3px solid var(--pst-color-border);

}
.bd-content .sd-tab-set>input:checked+label {
  border: none;
  transform: translateY(0);
  font-weight: 700;
  border-bottom: 4px solid var(--pst-color-secondary);
}
.bd-content .sd-tab-set>input:focus-visible+label {
  border: 2px outset var(--pst-color-secondary);
  transform: translateY(0);
}
.bd-content .sd-tab-set>label {
  border: none;
  background-color: transparent;
  font-weight: 500;
}

.card-title {
  font-size: 1.2rem;
  font-weight: bold;
}

.card-title svg {
  font-size: 2rem;
  vertical-align: bottom;
  margin-right: 5px;
}

/* This is gross, but necessary to meet accessibility requirements */
.headerlink {
  visibility: visible !important;
}
/* jupyter notebook output cells */
.bd-article .docutils .cell_output .output .highlight > pre:focus-visible{
  border: 2px outset var(--pst-color-secondary);
}

/* Copy button */
.bd-article .docutils .docutils .copybtn:focus-visible:after {
  /* border: 10px outset var(--pst-color-primary); */
    display: block;
    opacity: 1;
    visibility: visible;
}

/* Long autodoc module names wrap on prev/next links */
/* TODO: Should we extend this to the entire site? */
.prev-next-title {
  word-break: break-word;
}