ROOT_LOGGER_NAME = "autogen_core"
"""str: Logger name used for root logger"""

EVENT_LOGGER_NAME = "autogen_core.events"
"""str: Logger name used for structured event logging"""


TRACE_LOGGER_NAME = "autogen_core.trace"
"""str: Logger name used for developer intended trace logging. The content and format of this log should not be depended upon."""
